"""Unit tests for individual functions in reconciliation.py."""

import pytest
from collections_extended import RangeMap
from pymultirole_plugins.v1.schema import Annotation, Document, Term

from pyprocessors_reconciliation.reconciliation import (
    ReconciliationParameters,
    ReconciliationProcessor,
    ReconciliationType,
    WrappedTerm,
    by_label,
    by_lexicon,
    by_lexicon_or_label,
    group_annotations,
    has_knowledge,
    is_kill_or_white_label,
    is_suspicious,
    is_whitelist,
    labels_are_compatible,
    left_longest_match,
    mark_whitelisted,
    natural_order,
    one_match,
    person_varnames,
)

# ── helpers ───────────────────────────────────────────────────────────────────


def ann(start, end, label="ORG", text=None, terms=None):
    return Annotation(start=start, end=end, labelName=label, text=text, terms=terms)


def term(identifier, lexicon, preferred=None, properties=None):
    return Term(identifier=identifier, lexicon=lexicon, preferredForm=preferred or identifier, properties=properties)


def make_range_map(*annotations):
    rm = RangeMap()
    for a in annotations:
        rm[a.start : a.end] = a
    return rm


# ── WrappedTerm ───────────────────────────────────────────────────────────────


class TestWrappedTerm:
    def test_equal_same_id_and_status(self):
        t = term("id1", "lex", properties={"status": "linked"})
        assert WrappedTerm(t) == WrappedTerm(t)

    def test_not_equal_different_status(self):
        t1 = term("id1", "lex", properties={"status": "linked"})
        t2 = term("id1", "lex", properties={"status": "white"})
        assert WrappedTerm(t1) != WrappedTerm(t2)

    def test_equal_no_properties(self):
        t = term("id1", "lex")
        assert WrappedTerm(t) == WrappedTerm(t)

    def test_hash_consistent_for_equal(self):
        t = term("id1", "lex")
        assert hash(WrappedTerm(t)) == hash(WrappedTerm(t))

    def test_set_deduplication(self):
        t = term("id1", "lex")
        s = {WrappedTerm(t), WrappedTerm(t)}
        assert len(s) == 1

    def test_status_defaults_to_empty_without_properties(self):
        t = term("id1", "lex")
        wt = WrappedTerm(t)
        assert wt.status == ""


# ── has_knowledge ─────────────────────────────────────────────────────────────


class TestHasKnowledge:
    def test_with_terms(self):
        assert has_knowledge(ann(0, 5, terms=[term("id1", "lex")]))

    def test_none_terms(self):
        assert not has_knowledge(ann(0, 5, terms=None))

    def test_empty_terms(self):
        assert not has_knowledge(ann(0, 5, terms=[]))


# ── is_suspicious ─────────────────────────────────────────────────────────────


class TestIsSuspicious:
    def test_all_lowercase(self):
        assert is_suspicious(ann(0, 7, text="company")) is True

    def test_uppercase_first_word(self):
        assert is_suspicious(ann(0, 7, text="Company")) is False

    def test_multi_word_all_lower(self):
        assert is_suspicious(ann(0, 10, text="some thing")) is True

    def test_multi_word_one_upper(self):
        assert is_suspicious(ann(0, 11, text="some Thing")) is False

    def test_no_text_returns_false(self):
        assert is_suspicious(ann(0, 5)) is False

    def test_number_string(self):
        assert is_suspicious(ann(0, 3, text="123")) is True


# ── is_whitelist ──────────────────────────────────────────────────────────────


class TestIsWhitelist:
    def test_no_terms(self):
        assert is_whitelist(ann(0, 5)) is False

    def test_term_with_lowercase_w_status(self):
        a = ann(0, 5, terms=[term("id1", "lex", properties={"status": "w"})])
        assert is_whitelist(a) is True

    def test_term_with_uppercase_W_status(self):
        a = ann(0, 5, terms=[term("id1", "lex", properties={"status": "W"})])
        assert is_whitelist(a) is True

    def test_term_with_non_w_status(self):
        a = ann(0, 5, terms=[term("id1", "lex", properties={"status": "linked"})])
        assert is_whitelist(a) is False

    def test_term_with_no_properties(self):
        assert is_whitelist(ann(0, 5, terms=[term("id1", "lex")])) is False


# ── person_varnames ───────────────────────────────────────────────────────────


class TestPersonVarnames:
    def test_non_person_label_returns_none(self):
        assert person_varnames(ann(0, 7, label="ORG", text="Company"), "Company") is None

    def test_single_word_name(self):
        a = ann(0, 4, label="person", text="Jean")
        assert person_varnames(a, "Jean") == ["Jean"]

    def test_two_word_name_contains_all_variants(self):
        a = ann(0, 11, label="person", text="Jean Dupont")
        result = person_varnames(a, "Jean Dupont")
        assert "Jean Dupont" in result
        assert "Dupont" in result
        assert "Jean" in result

    def test_three_word_name_variants(self):
        a = ann(0, 16, label="person", text="Jean Paul Dupont")
        result = person_varnames(a, "Jean Paul Dupont")
        assert "Jean Paul Dupont" in result
        assert "Paul Dupont" in result
        assert "Dupont" in result
        assert "Jean" in result
        assert "Jean Paul" in result

    def test_uses_text_slice_when_annotation_has_no_text(self):
        text = "Hello Jean Dupont end"
        a = Annotation(start=6, end=17, labelName="person")
        result = person_varnames(a, text)
        assert result is not None
        assert "Jean Dupont" in result


# ── key functions ─────────────────────────────────────────────────────────────


class TestKeyFunctions:
    def test_by_lexicon_with_terms(self):
        assert by_lexicon(ann(0, 5, terms=[term("id1", "wikidata")])) == "wikidata"

    def test_by_lexicon_no_terms(self):
        assert by_lexicon(ann(0, 5)) == ""

    def test_by_lexicon_or_label_with_terms(self):
        a = ann(0, 5, label="ORG", terms=[term("id1", "wikidata")])
        assert by_lexicon_or_label(a) == "wikidata"

    def test_by_lexicon_or_label_no_terms(self):
        assert by_lexicon_or_label(ann(0, 5, label="ORG")) == "ORG"

    def test_by_label(self):
        assert by_label(ann(0, 5, label="PER")) == "PER"

    def test_left_longest_match(self):
        assert left_longest_match(ann(2, 7)) == (5, -2)

    def test_natural_order(self):
        assert natural_order(ann(3, 8)) == (-3, 5)


# ── one_match ─────────────────────────────────────────────────────────────────


class TestOneMatch:
    def test_none_matches_returns_false_none(self):
        perfect, match = one_match(ann(0, 5), None)
        assert perfect is False
        assert match is None

    def test_perfect_match(self):
        a = ann(0, 5, label="ORG")
        b = ann(0, 5, label="ORG")
        perfect, match = one_match(a, make_range_map(b))
        assert perfect is True
        assert match is b

    def test_imperfect_match_larger_span(self):
        a = ann(2, 5, label="ORG")
        b = ann(0, 7, label="ORG")
        perfect, match = one_match(a, make_range_map(b))
        assert perfect is False
        assert match is b


# ── label compatibility ───────────────────────────────────────────────────────


class TestLabelCompatibility:
    @pytest.fixture
    def params(self):
        return ReconciliationParameters(kill_label="kill", white_label="white")

    def test_same_label_compatible(self, params):
        assert labels_are_compatible("ORG", "ORG", params) is True

    def test_kill_label_is_compatible(self, params):
        assert labels_are_compatible("kill", "ORG", params) is True

    def test_white_label_is_compatible(self, params):
        assert labels_are_compatible("ORG", "white", params) is True

    def test_unrelated_labels_incompatible(self, params):
        assert labels_are_compatible("ORG", "PER", params) is False

    def test_is_kill_label(self, params):
        assert is_kill_or_white_label("kill", params) is True

    def test_is_white_label(self, params):
        assert is_kill_or_white_label("white", params) is True

    def test_unrelated_label_is_neither(self, params):
        assert is_kill_or_white_label("ORG", params) is False


# ── ReconciliationParameters defaults ─────────────────────────────────────────


class TestReconciliationParametersDefaults:
    def test_defaults(self):
        p = ReconciliationParameters()
        assert p.type == ReconciliationType.linker
        assert p.kill_label is None
        assert p.white_label is None
        assert p.whitelisted_lexicons is None
        assert p.person_label is None
        assert p.remove_suspicious is True
        assert p.resolve_lastnames is False


# ── group_annotations ─────────────────────────────────────────────────────────


class TestGroupAnnotations:
    def test_non_overlapping_annotations_in_same_group(self):
        t = term("id1", "lex")
        a1 = ann(0, 5, terms=[t])
        a2 = ann(10, 15, terms=[t])
        groups = group_annotations([a1, a2], by_lexicon)
        assert len(groups["lex"]) == 2

    def test_same_span_different_terms_merged(self):
        t1 = term("id1", "lex")
        t2 = term("id2", "lex")
        a1 = ann(0, 5, terms=[t1])
        a2 = ann(0, 5, terms=[t2])
        groups = group_annotations([a1, a2], by_lexicon)
        merged = groups["lex"][0]
        assert len(merged.terms) == 2

    def test_same_span_duplicate_term_deduplicated(self):
        t = term("id1", "lex")
        a1 = ann(0, 5, terms=[t])
        a2 = ann(0, 5, terms=[t])
        groups = group_annotations([a1, a2], by_lexicon)
        merged = groups["lex"][0]
        assert len(merged.terms) == 1

    def test_empty_input(self):
        groups = group_annotations([], by_lexicon)
        assert len(groups) == 0

    def test_annotations_split_into_groups_by_lexicon(self):
        a1 = ann(0, 5, terms=[term("id1", "wiki")])
        a2 = ann(0, 5, terms=[term("id2", "kb")])
        groups = group_annotations([a1, a2], by_lexicon)
        assert "wiki" in groups
        assert "kb" in groups


# ── mark_whitelisted ──────────────────────────────────────────────────────────


class TestMarkWhitelisted:
    def test_passthrough_when_no_config(self):
        t = term("id1", "lex")
        annotations = [ann(0, 5, terms=[t]), ann(10, 15)]
        result = mark_whitelisted(annotations, ReconciliationParameters())
        assert result == annotations

    def test_white_label_clears_terms(self):
        t = term("id1", "lex")
        a = Annotation(start=0, end=5, labelName="white", terms=[t])
        params = ReconciliationParameters(white_label="white")
        result = mark_whitelisted([a], params)
        assert len(result) == 1
        assert result[0].terms is None

    def test_whitelisted_lexicon_adds_term_free_duplicate(self):
        t = term("id1", "mylex")
        a = ann(0, 5, label="ORG", terms=[t])
        params = ReconciliationParameters(whitelisted_lexicons=["mylex"])
        result = mark_whitelisted([a], params)
        assert len(result) == 2
        assert result[0].terms is not None  # original kept
        assert result[1].terms is None  # copy without terms

    def test_annotation_without_terms_always_passed_through(self):
        a = ann(0, 5)
        params = ReconciliationParameters(white_label="white", whitelisted_lexicons=["mylex"])
        result = mark_whitelisted([a], params)
        assert len(result) == 1


# ── process: edge cases ───────────────────────────────────────────────────────


class TestProcessEdgeCases:
    @pytest.fixture
    def processor(self):
        return ReconciliationProcessor()

    def test_document_without_annotations_returned_unchanged(self, processor):
        doc = Document(text="Hello world")
        result = processor.process([doc], ReconciliationParameters())
        assert result[0].annotations is None

    def test_sentence_annotations_stripped_before_reconciliation(self, processor):
        model_ann = Annotation(start=0, end=5, labelName="ORG", text="Apple")
        sentence = Annotation(start=0, end=11, labelName="sentence", text="Apple world")
        doc = Document(text="Apple world", annotations=[model_ann, sentence])
        result = processor.process([doc], ReconciliationParameters())
        labels = [a.labelName for a in result[0].annotations]
        assert "sentence" not in labels
        assert "ORG" in labels

    def test_suspicious_annotation_removed_by_default(self, processor):
        a = Annotation(start=0, end=7, labelName="ORG", text="company")
        doc = Document(text="company", annotations=[a])
        result = processor.process([doc], ReconciliationParameters(remove_suspicious=True))
        assert result[0].annotations == []

    def test_suspicious_annotation_kept_when_flag_disabled(self, processor):
        a = Annotation(start=0, end=7, labelName="ORG", text="company")
        doc = Document(text="company", annotations=[a])
        result = processor.process([doc], ReconciliationParameters(remove_suspicious=False))
        assert len(result[0].annotations) == 1

    def test_kill_label_removes_matching_model_annotation(self, processor):
        model_ann = Annotation(start=0, end=5, labelName="ORG", text="Apple")
        killer = Annotation(start=0, end=5, labelName="kill", text="Apple", terms=[term("id1", "blocklist")])
        doc = Document(text="Apple falls", annotations=[model_ann, killer])
        result = processor.process([doc], ReconciliationParameters(kill_label="kill"))
        labels = [a.labelName for a in result[0].annotations]
        assert "ORG" not in labels

    def test_kb_annotation_enriches_matching_model_annotation(self, processor):
        kb = Annotation(start=0, end=5, labelName="ORG", text="Apple", terms=[term("Q312", "wikidata", "Apple Inc.")])
        model_ann = Annotation(start=0, end=5, labelName="ORG", text="Apple")
        doc = Document(text="Apple rules", annotations=[model_ann, kb])
        result = processor.process([doc], ReconciliationParameters())
        enriched = result[0].annotations[0]
        assert enriched.terms is not None
        assert any(t.lexicon == "wikidata" for t in enriched.terms)

    def test_multiple_documents_processed_independently(self, processor):
        doc1 = Document(text="Apple", annotations=[Annotation(start=0, end=5, labelName="ORG", text="Apple")])
        doc2 = Document(text="Hello world")
        result = processor.process([doc1, doc2], ReconciliationParameters(remove_suspicious=False))
        assert len(result[0].annotations) == 1
        assert result[1].annotations is None



import copy
from collections import OrderedDict, defaultdict
from collections.abc import Iterable
from enum import StrEnum
from itertools import groupby
from typing import cast

from collections_extended import RangeMap
from log_with_context import Logger, add_logging_context
from pydantic import BaseModel, Field
from pymultirole_plugins.v1.processor import ProcessorBase, ProcessorParameters
from pymultirole_plugins.v1.schema import Annotation, Document

logger = Logger("pymultirole")


class WrappedTerm:
    def __init__(self, term):
        self.term = term
        self.status = term.properties.get("status", "") if term.properties else ""

    def __eq__(self, other):
        return self.term.identifier == other.term.identifier and self.status == other.status

    def __hash__(self):
        return hash((self.term.identifier, self.status))


class ReconciliationType(StrEnum):
    linker = "linker"


class ReconciliationParameters(ProcessorParameters):
    type: ReconciliationType = Field(
        ReconciliationType.linker,
        description="""Type of consolidation, use<br />
    <li>**linker** to retain only known entities (extracted by a model)<br />""",
        json_schema_extra={"extra": "internal,advanced"},
    )
    kill_label: str | None = Field(
        None, description="Label name of the kill list", json_schema_extra={"extra": "label,advanced"}
    )
    white_label: str | None = Field(
        None, description="Label name of the white list", json_schema_extra={"extra": "label,advanced"}
    )
    whitelisted_lexicons: list[str] = Field(
        None,
        description="Lexicons to be considered as taking precedence over models",
        json_schema_extra={"extra": "lexicon,advanced"},
    )
    person_label: str | None = Field(
        None,
        description="Label name of the person to apply the lastname resolution",
        json_schema_extra={"extra": "label,advanced"},
    )
    remove_suspicious: bool = Field(
        True,
        description="Remove suspicious annotations extracted by the model"
        " (numbers, percentages, phrases without uppercase words)",
        json_schema_extra={"extra": "advanced"},
    )
    resolve_lastnames: bool = Field(
        False,
        description="Try to resolve isolated family names and firstnames if they have been seen before in the document",
        json_schema_extra={"extra": "advanced"},
    )


class ReconciliationProcessor(ProcessorBase):
    """Reconciliation processor ."""

    def process(self, documents: list[Document], parameters: ProcessorParameters) -> list[Document]:  # noqa: C901
        params: ReconciliationParameters = cast(ReconciliationParameters, parameters)
        for document in documents:
            with add_logging_context(docid=document.identifier):
                if document.annotations:
                    annotations = [a for a in document.annotations if a.labelName != "sentence"]
                    annotations = mark_whitelisted(annotations, params)
                    ann_groups = group_annotations(annotations, keyfunc=by_lexicon)
                    # Consolidate & links against KB and Wikidata
                    if params.type == ReconciliationType.linker:
                        conso_anns = consolidate_linker(document.text, ann_groups, params)
                    document.annotations = conso_anns
        return documents

    @classmethod
    def get_model(cls) -> type[BaseModel]:
        return ReconciliationParameters


def mark_whitelisted(annotations, params: ReconciliationParameters):
    new_annotations = []
    white_label = params.white_label
    white_lexicons = params.whitelisted_lexicons or []
    for a in annotations:
        new_annotations.append(a)
        if has_knowledge(a):
            if a.labelName == white_label:  # Consider whitelisted terms as entities coming from the model
                a.terms = None
            elif a.terms[0].lexicon in white_lexicons:
                new_a = copy.deepcopy(a)
                new_a.terms = None
                new_annotations.append(new_a)
    return new_annotations


def consolidate_linker(text, ann_groups, params: ReconciliationParameters):
    conso_anns = []
    kill_names = defaultdict(set)
    kb_names = defaultdict(set)
    white_names = defaultdict(set)
    for k, v in ann_groups.items():
        if k != "":
            for a in v.values():
                if a.labelName == params.kill_label:
                    kill_names[k].add(a.labelName)
                elif a.labelName == params.white_label:
                    white_names[k].add(a.labelName)
                else:
                    kb_names[k].add(a.labelName)

    for model_r in ann_groups[""].ranges():
        model_ann = model_r.value

        if params.remove_suspicious and is_suspicious(model_ann):
            logger.warning("Kill suspicious annotation")
            logger.warning(f"=> {model_ann}")
            continue
        if kill_names:
            kill_r = annotation_in_group(model_ann, ann_groups, params, kill_names)
            perfect, kill_match = one_match(model_ann, kill_r)
            if perfect and kill_match:
                logger.warning("Kill annotation")
                logger.warning(f"=> {model_ann}")
                continue

        kb_r = annotation_in_group(model_ann, ann_groups, params, kb_names)
        perfect, kb_match = one_match(model_ann, kb_r)
        if kb_match:
            if perfect:
                if model_ann.labelName == kb_match.labelName or model_ann.labelName == params.white_label:
                    model_ann.labelName = kb_match.labelName
                    model_ann.label = kb_match.label
                    model_ann.terms = model_ann.terms or []
                    model_ann.terms.extend(kb_match.terms)
                else:
                    logger.warning("Found wrong label annotation in KB")
                    logger.warning(f"=> {model_ann}")
                    logger.warning("and")
                    logger.warning(f" -{kb_match}")
            else:
                logger.warning("Found larger annotation in KB")
                logger.warning(f"=> {model_ann}")
                logger.warning("and")
                logger.warning(f" -{kb_match}")
        elif kb_r and len(kb_r) > 1:
            logger.warning("Found overlapping annotations in KB")
            logger.warning(f"=> {model_ann}")
            logger.warning("and")
            for r in kb_r.values():
                logger.warning(f" -{r}")

        # wiki_r = annotation_in_group(model_ann, ann_groups, [params.wikidata_label])
        # perfect, wiki_match = one_match(model_ann, wiki_r)
        # if wiki_match:
        #     if validate_wiki_type(wiki_match, gname):
        #         if perfect:
        #             model_ann.terms = model_ann.terms or []
        #             wiki_match.terms[0].properties.pop("fingerprint", None)
        #             model_ann.terms.extend(wiki_match.terms)
        #         else:
        #             logger.warning("Found larger annotation in Wikidata")
        #             logger.warning(f"=> {model_ann}")
        #             logger.warning("and")
        #             logger.warning(f" -{wiki_match}")
        # elif wiki_r and len(wiki_r) > 1:
        #     logger.warning("Found overlapping annotations in Wikidata")
        #     logger.warning(f"=> {model_ann}")
        #     logger.warning("and")
        #     for r in wiki_r.values():
        #         logger.warning(f" -{r}")
        conso_anns.append(model_ann)
    sorted_annotations = sorted(
        conso_anns,
        key=natural_order,
        reverse=True,
    )
    seen_names = defaultdict(set)
    if params.resolve_lastnames:
        for ann in sorted_annotations:
            if ann.labelName == "person":
                lastnames = person_varnames(ann, text)
                if lastnames is not None:
                    for composed_name in lastnames:
                        if len(lastnames) > 1:
                            if has_knowledge(ann):
                                for t in ann.terms:
                                    seen_names[composed_name].add(WrappedTerm(t))
                            else:
                                seen_names[composed_name].add(None)
                        else:
                            if composed_name in seen_names:
                                ann.terms = [wt.term for wt in seen_names[composed_name] if wt is not None]
                                break

    return sorted_annotations


def group_annotations(annotations, keyfunc):
    groups = defaultdict(RangeMap)
    sorted_annotations = sorted(annotations, key=keyfunc)
    for k, g in groupby(sorted_annotations, keyfunc):
        sorted_group = sorted(g, key=left_longest_match, reverse=True)
        for a in sorted_group:
            addit = True
            if a.start in groups[k] and a.end - 1 in groups[k]:
                b = groups[k][a.start]
                if a.start - b.start == 0 and a.end - b.end == 0:
                    terms = {WrappedTerm(t) for t in b.terms} if b.terms else set()
                    if a.terms:
                        terms.update({WrappedTerm(t) for t in a.terms})
                    b.terms = [t.term for t in terms]
                    addit = False
            if addit:
                groups[k][a.start : a.end] = a
    return groups


def left_longest_match(a: Annotation):
    return a.end - a.start, -a.start


def natural_order(a: Annotation):
    return -a.start, a.end - a.start


def has_knowledge(a: Annotation):
    return a.terms is not None and a.terms


def is_whitelist(a: Annotation):
    if has_knowledge(a):
        for term in a.terms:
            props = term.properties or {}
            status = props.get("status", "")
            if "w" in status.lower():
                return True
    return False


def person_varnames(a: Annotation, text):
    if "person" in a.labelName:
        atext = a.text or text[a.start : a.end]
        words = atext.split()
        if words is not None:
            variant_names = OrderedDict.fromkeys([" ".join(words[i:]) for i in range(len(words))])
            if len(words) > 1:
                variant_names2 = OrderedDict.fromkeys([" ".join(words[0 : i - 1]) for i in range(2, len(words) + 1)])
                variant_names.update(variant_names2)
            return list(variant_names.keys())
    return None


def by_lexicon(a: Annotation):
    if a.terms:
        lex = a.terms[0].lexicon
        return lex
    else:
        return ""


def by_lexicon_or_label(a: Annotation):
    if a.terms:
        lex = a.terms[0].lexicon
        return lex
    else:
        return a.labelName


def by_label(a: Annotation):
    return a.labelName


def one_match(a: Annotation, matches: RangeMap):
    match = None
    perfect = False
    if matches and len(matches) >= 1:
        for match in matches.values():
            perfect = a.start == match.start and a.end == match.end
            if perfect:
                break
    return perfect, match


def is_suspicious(a: Annotation):
    suspicious = False
    if a.text:
        words = a.text.split()
        has_upper = any(w[0].isupper() for w in words)
        suspicious = not has_upper
    return suspicious


# noqa: W503
# def annotation_in_group(
#         a: Annotation, ann_groups: Dict[str, RangeMap], gnames: List[str] = None
# ):
#     gname = by_lexicon_or_label(a)
#     if gname in gnames:
#         gnames = [gname]
#     for gname in gnames:
#         if (
#                 gname in ann_groups
#                 and a.start in ann_groups[gname]
#                 or a.end in ann_groups[gname]
#         ):
#             ga = ann_groups[gname][a.start: a.end]
#             return ga
#     return None


def is_kill_or_white_label(label: str, params: ReconciliationParameters):
    return label == params.white_label or label == params.kill_label


def labels_are_compatible(a: str, b: str, params: ReconciliationParameters):
    return a == b or is_kill_or_white_label(a, params) or is_kill_or_white_label(b, params)


def annotation_in_group(
    a: Annotation,
    ann_groups: dict[str, RangeMap],
    params: ReconciliationParameters,
    gnames: dict[str, Iterable[str]] = None,
):
    for gkey, glabels in gnames.items():
        if any(labels_are_compatible(a.labelName, glabel, params) for glabel in glabels) and (
            gkey in ann_groups and (a.start in ann_groups[gkey] or a.end - 1 in ann_groups[gkey])
        ):
            ga = ann_groups[gkey][a.start : a.end]
            return ga
    return None
